'''
APDT: Processing
--------------------------

Provide useful data processing tools for further usage.

Check https://github.com/Zhiyuan-Wu/apdt for more information.
'''
from apdt.proc.interpolate import spatial_interplot, temporal_interpolate
from apdt.proc.utils import linear_normalize, time_stamp_feature